import LineProtocolGenerator as lp
import time
import urllib.parse
import urllib.request
import sys
import random

# Simulation parameters
TICK_TIME = 1
DEFAULT_REQUEST_RATE_INCREMENT = 5
SIMULATION_TIME_SEC = 180

# CLMC parameters
INFLUX_DB_URL = 'http://192.168.50.10:8086'

# Simulator for services
class sim:
    def __init__(self, influx_url):
        # We don't need this as the db is CLMC metrics
        self.influx_db = 'CLMCMetrics'
        self.influx_url = influx_url
        # Teardown DB from previous sim and bring it back up
        self._deleteDB()
        self._createDB()


    def run(self, simulation_length_seconds):
        start_time = time.time()-SIMULATION_TIME_SEC
        sim_time = start_time

        ip_endpoints = [{'agent_url': 'http://192.168.50.11:8186', 'location': 'DC1', 'cpu': 16,
                        'mem': '8GB', 'storage': '1TB', 'request_queue': 0, 'request_arrival_rate': 0},
                        {'agent_url': 'http://192.168.50.12:8186', 'location': 'DC2', 'cpu': 2, 
                        'mem': '8GB', 'storage': '1TB', 'request_queue': 0, 'request_arrival_rate': 0}
                        ]

        # Simulate configuration of the ipendpoints
        # endpoint state->mu, sigma, secs normal distribution
        config_delay_dist = {"placed": [5, 0.68], "booted": [10, 0.68],"connected": [10, 0.68]}

        # Place the endpoints
        max_delay = 0
        for ip_endpoint in ip_endpoints:
            delay_time = random.normalvariate(config_delay_dist['placed'][0], config_delay_dist['placed'][0]*config_delay_dist['placed'][1])
            
         #   print('sim_time: ', sim_time)            
         #   print('delay_time: ', delay_time)
         #   timestamp = sim_time+delay_time
         #   print('timestamp: ', timestamp)
         #   ns_time = int(timestamp*1000000)
         #   print('ns_time: ', ns_time)

            self._sendInfluxData(ip_endpoint['agent_url'], lp.generate_vm_config('placed', ip_endpoint['cpu'], ip_endpoint['mem'], ip_endpoint['storage'], sim_time+delay_time)) 
            if delay_time > max_delay:
                max_delay = delay_time

        sim_time +=max_delay
        max_delay = 0        
        # Boot the endpoints
        for ip_endpoint in ip_endpoints:
            delay_time = random.normalvariate(config_delay_dist['booted'][0], config_delay_dist['booted'][0]*config_delay_dist['booted'][1])
            self._sendInfluxData(ip_endpoint['agent_url'], lp.generate_vm_config('booted', ip_endpoint['cpu'], ip_endpoint['mem'], ip_endpoint['storage'], sim_time+delay_time))
            if delay_time > max_delay:
                max_delay = delay_time            

        sim_time +=max_delay
        max_delay = 0     
        # Connect the endpoints
        for ip_endpoint in ip_endpoints:
            delay_time = random.normalvariate(config_delay_dist['connected'][0], config_delay_dist['connected'][0]*config_delay_dist['connected'][1])
            self._sendInfluxData(ip_endpoint['agent_url'], lp.generate_vm_config('connected', ip_endpoint['cpu'], ip_endpoint['mem'], ip_endpoint['storage'], sim_time+delay_time))
            if delay_time > max_delay:
                max_delay = delay_time

        sim_time +=max_delay
   
        request_arrival_rate_inc = DEFAULT_REQUEST_RATE_INCREMENT
        request_queue = 0
        for i in range(simulation_length_seconds):        
            for ip_endpoint in ip_endpoints:
                request_processing_time = 0
                cpu_time_available = 0
                requests_processed = 0
                max_requests_processed = 0
                cpu_active_time = 0
                cpu_idle_time = 0
                cpu_usage = 0
                cpu_load_time = 0
                avg_response_time = 0
                peak_response_time = 0

                # linear inc to arrival rate
                ip_endpoint['request_arrival_rate'] += request_arrival_rate_inc
                # add new requests to the queue
                ip_endpoint['request_queue'] += ip_endpoint['request_arrival_rate']

                # time to process one request (mS) in the current second
                request_processing_time = int(random.normalvariate(50, 50*0.95))
                if request_processing_time <= 10:
                    request_processing_time = 10
                # amount of cpu time (mS) per tick
                cpu_time_available = ip_endpoint['cpu']*TICK_TIME*1000
                max_requests_processed = cpu_time_available/request_processing_time
                # calc how many requests processed
                if ip_endpoint['request_queue'] <= max_requests_processed:
                    # processed all of the requests
                    requests_processed = ip_endpoint['request_queue']
                else:
                    # processed the maxmum number of requests
                    requests_processed = max_requests_processed
                # calculate cpu usage
                cpu_active_time = int(requests_processed*request_processing_time)
                cpu_idle_time = int(cpu_time_available-cpu_active_time)
                cpu_usage = cpu_active_time/cpu_time_available
                self._sendInfluxData(ip_endpoint['agent_url'], lp.generate_cpu_report(cpu_usage, cpu_active_time, cpu_idle_time, sim_time))

                # calc network usage metrics with no constraints. 
                bytes_sent = 1024*requests_processed
                bytes_rec = 32*requests_processed
                self._sendInfluxData(ip_endpoint['agent_url'], lp.generate_network_report(bytes_rec, bytes_sent, sim_time))                

                # time to process all of the requests in the queue
                peak_response_time = ip_endpoint['request_queue']*request_processing_time/ip_endpoint['cpu']
                # mid-range 
                avg_response_time = (peak_response_time+request_processing_time)/2
                self._sendInfluxData(ip_endpoint['agent_url'], lp.generate_mpegdash_report('https://Netflix.com/scream', ip_endpoint['request_arrival_rate'], avg_response_time, peak_response_time, sim_time))

                # remove requests processed off the queue
                ip_endpoint['request_queue'] -= int(requests_processed)            

            sim_time += TICK_TIME
        end_time = sim_time
        print("Simulation Finished. Start time {0}. End time {1}. Total time {2}".format(start_time,end_time,end_time-start_time))

    def _createDB(self):
        self._sendInfluxQuery(self.influx_url, 'CREATE DATABASE ' + self.influx_db)


    def _deleteDB(self):
        self._sendInfluxQuery(self.influx_url, 'DROP DATABASE ' + self.influx_db)


    def _sendInfluxQuery(self, url, query):
        query = urllib.parse.urlencode({'q': query})
        query = query.encode('ascii')
        req = urllib.request.Request(url + '/query ', query)
        urllib.request.urlopen(req)

    def _sendInfluxData(self, url, data):
        data = data.encode()
        header = {'Content-Type': 'application/octet-stream'}
        req = urllib.request.Request(url + '/write?db=' + self.influx_db, data, header)
        urllib.request.urlopen(req)  


simulator = sim(INFLUX_DB_URL)
simulator.run(SIMULATION_TIME_SEC)

